<html>

  <head></head>

  <body>

    <userSplice>
      <h2>
	<ifLoggedIn>
	  <loggedInUser/> is logged in
	</ifLoggedIn>
	
	<ifLoggedOut>
	  You are not logged in
	</ifLoggedOut>
      </h2>

      <h3>loggedInUser: <loggedInUser/></h3>

      <p>UserID <userId/></p>

      <p>UserLogin <userLogin/></p>
      
      <p>UserEmail <userEmail/></p>

      <p>UserActive <userActive/></p>
      
      <p>UserLoginCount <userLoginCount/></p>
      
      <p>UserFailedCount <userFailedCount/></p>
      
      <p>UserLoginAt <userLoginAt/></p>
      
      <p>UserLastLoginAt <userLastLoginAt/></p>
      
      <p>UserSuspendedAt <userSuspendedAt/></p>
      
      <p>UserLoginIP <userLoginIP/></p>
      
      <p>UserLastLoginIP <userLastLoginIP/></p>
      
      <p>UserIfActive <userIfActive/></p>
      
      <p>userIfSuspended <userIfSuspended/></p>
      
    </userSplice>
    
  </body>
  
</html>
